#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <unistd.h>
#include <time.h>

#include "crypt.h"
#include "rabin.h"
#include "umac.h"


#include "th_assert.h"
#include "Timer.h"
#include "libbyz.h"

#include "Statistics.h"

#include "simple.h"



const int UMAC_size = 8;
const int UNonce_size = sizeof(long long);
const int MAC_size =   UMAC_size + UNonce_size;

const int Key_size = 16;
const int Key_size_u = Key_size/sizeof(unsigned);
unsigned kout[Key_size_u];

umac_ctx_t ctx_in=0;
umac_ctx_t ctx_out;//=umac_new((char*));

long long umac_nonce = 0;

bool verify_mac(const char *src, unsigned src_len,
                           const char *mac, const char *unonce, umac_ctx_t ctx) {
  // Do not accept MACs sent with uninitialized keys.
  if (ctx == 0) return false;

  char tag[20];
  umac(ctx, (char *)src, src_len, tag, (char *)unonce);
  umac_reset(ctx);
  return !memcmp(tag, mac, UMAC_size);
}


void gen_mac(const char *src, unsigned src_len,
			char *dst, const char *unonce, umac_ctx_t ctx) {
  umac(ctx, (char *)src, src_len, dst, (char *)unonce);
  umac_reset(ctx);
}







int main(int argc, char **argv) {
   
  char pk1[1024] = "d3bf9ada150474e93d21a4818ccf40e97df94f565c0528973a7799fc3e9ee69e0561fff15631850e2c5b8f9accee851cfc170cd0193052d4f75dfee18ab1d24b";
  
  char pk2[1024] = "e7b8885fa504355a686140181ae956e726e490ac2f905e52a78bea2ef16acef31788b827f35f0de1343766e6f2cbe44f436d7e5eceeb67791ccd296422fb50ef";
    
  char pk3[1024] = "bfaa873efc926cb91646a89e45f96582041e3eed35cde0ef60b5c006cfad883781ee807411b0df3c74dc3ebbbce59c21d67711c83ecf596357c23dba33da338fb5577179a3b6188c59590aa1301eb852c0e14fa9225c0b377fee944eb9fa110ad7a316269e4b13b153887426a347c7c3c5feb1e3107bac4c6e29327b3343c405";
  
//  printf("Calling rabin_priv...\n"); 
//  rabin_priv *priv_key = new rabin_priv(bigint(pk1,16), bigint(pk2,16));
//  printf("calling rabin_pub...\n");
//  rabin_pub *pub_key = new rabin_pub(bigint(pk3,16));


  printf("Calling rabin_priv...\n");
  rsa_priv *priv_key = new rsa_priv(bigint(pk1,16), bigint(pk2,16));
  printf("calling rabin_pub...\n");
  rsa_pub *pub_key = new rsa_pub(bigint(pk3,16));

  bigint sig;
  int iter = 10000;
  int incr = 100;

  const str msg = "Hello world, how are you doing today?aaferwh";

  long t1 = time(0);
  for (int i = 0; i <iter; i++)
    sig = priv_key->sign(msg);
  long t2 = time(0);

  float g = t2, h = t1;
  float f = g-h;
  f = f/iter;

  printf("Time to sign        %d times: %d seconds, or %f seconds per signature\n", iter, t2-t1, f);

  t1 = time(0);
  iter *= incr;
  for (int i = 0; i < iter; i++){
    bool res = pub_key->verify(msg, sig);
    if (!res){
      printf("failed to verify\n");
      exit(-1);
    }
  }
  t2 = time(0);
  g = t2;
  h = t1;
  f = g-h;
  f = f/iter;
  printf("Time to verify    %d times: %d seconds, or %f seconds per signature\n", iter, t2-t1, f);


  for (int i = 0; i < Key_size_u; i++)
    kout[i]=0;
  
  char dst[MAC_size];  // holds at least MAC_size bytes
  umac_ctx_t ctx= umac_new((char*)kout);  // points to an initialized umac context
  unsigned src_len=35;  // number of bytes to be authenticated
  char src[src_len];
  for(int i = 0; i < src_len; i++)
    src[i] = 'a'+i;

  iter *=incr;
  t1 = time(0);
  for (int i = 0; i < iter; i++)
    gen_mac(src, src_len, dst, (char*)&umac_nonce, ctx);
  t2 = time(0);
  f = t2-t1;
  f = f/iter;
  
  printf("Time to generat %d macs:  %d seconds, or %f seconds per mac\n", iter, t2-t1, f);


  t1 = time(0);
  for (int i = 0; i < iter; i++){
    bool res = verify_mac(src, src_len, dst, (char*)&umac_nonce, ctx);
    //    if (!res){
    // printf("verify failed\n");
    // exit(-1);
    //   }
  }
  t2 = time(0);
  f = t2-t1;
  f = f/iter;
  printf("Time to verify  %d macs:  %d seconds, or %f seconds per mac\n", iter, t2 -t1, f);



  
}
  
