// libbyz-speculative
#include <string.h>
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <sys/types.h>  
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h> //for non_blocking socket
#include <errno.h>

#include "Array.h"
#include "Array.t"
#include "th_assert.h"
#include "Message_tags.h"
#include "ITimer.h"
#include "Request.h"
#include "Pre_prepare.h"
#include "Prepare.h"
#include "Commit.h"
#include "Checkpoint.h"
#include "New_key.h"
#include "Status.h"
#include "Fetch.h"
#include "Data.h"
#include "Meta_data.h"
#include "Meta_data_d.h"
#include "View_change.h"
#include "View_change_ack.h"
#include "New_view.h"
#include "Principal.h"
#include "Prepared_cert.h"
#include "Reply.h"
#include "K_max.h"
#include "Verifier_thread.h"
#include "Replica.h"
#include "Statistics.h"
#include "State_defs.h"
#include "attacks.h"

// Global replica object.
Replica *replica;

Verifier_thread *verifier;

// Force template instantiation
#include "Certificate.t"
template class Certificate<Commit>;
template class Certificate<Checkpoint>;
template class Certificate<Reply>;

#include "Log.t"
template class Log<Prepared_cert>;
template class Log< Certificate<Commit> >;
template class Log< Certificate<Checkpoint> >;

#include "Set.t"
template class Set<Checkpoint>;

template <class T> 
void Replica::retransmit(T *m, Time &cur,  
			 Time *tsent, Principal *p) {
  // re-authenticate.
  m->re_authenticate(p);
 //    printf("RET: %s to %d \n", m->stag(), p->pid());
 // Retransmit message
 send(m, p->pid());
}


Replica::Replica(FILE *config_file, FILE *config_priv, char *mem, int nbytes,  int _byz_pre_prepare_delay,
		 bool _small_batches, long int _exec_command_delay) :
  Node(config_file, config_priv), rqueue(), ro_rqueue(),
  plog(max_out), clog(max_out), elog(max_out*2,0), sset(n()), 
  replies(mem, nbytes, num_principals), state(this, mem, nbytes), 
  vi(node_id, 0) {


//static address table

  addresses = (char**)malloc(num_replicas*num_replicas*sizeof(char*));
  for(int j=0; j<num_replicas*num_replicas; j++){
    addresses[j] = (char*)malloc(50*sizeof(char));
  }

//static port table

  ports = (int*)malloc(num_replicas*num_replicas*sizeof(int));


  // Fail if node is not a replica.
  if (!is_replica(id())) th_fail("Node is not a replica");

  // The following two lines implement the 'slow' primary
  byz_pre_prepare_delay = _byz_pre_prepare_delay; 
  small_batches = _small_batches;
  exec_command_delay = _exec_command_delay;

  verifier = NULL;
  pthread_t     thread;
  int           rc;


//  printf("port number (from replica) %d\n",principals[id()]->address()->sin_port);
  verifier = new Verifier_thread(principals[id()]->address()->sin_port);
  rc = pthread_create(&thread, NULL, Verifier_thread_startup, verifier);
  if (rc) {
    printf("Failed to create a thread\n");
    exit(1);
  }

#ifdef FAIRNESS_ADAPTIVE
  max_rqueue_size = 0;
  executed_snapshot = (Seqno*)malloc(num_principals*sizeof(Seqno));
  for (int j=0; j< num_principals ; j++)
    executed_snapshot[j]=-1;
  fairness_bound = fairness_multiplier*num_principals;
#endif

#ifdef THROUGHPUT_ADAPTIVE
  last_cp_time = 0;
  req_throughput = req_throughput_init; 
  req_throughput_increment = req_throughput_increment_init;
  first_checkpoint_after_view_change = false;
  highest_throughput_ever = 0;
  vc_already_triggered = false;
  time_to_increment = false;
  checkpoints_in_new_view = 0;

  last_throughput_of_replica = (float*)malloc(num_replicas*sizeof(float));

  for(int j = 0; j < num_replicas; j++)
    last_throughput_of_replica[j] = req_throughput_init;
#endif

  seqno = 0;
  last_stable = 0;
  low_bound = 0; 

  last_prepared = 0;
#ifdef SPECULATIVE
  last_committed = 0;
#else
  last_executed = 0;
#endif
  last_tentative_execute = 0;
#ifdef USE_GETTIMEOFDAY 
  last_status.tv_sec=0;
  last_status.tv_nsec=0;
#else
  last_status = 0;
#endif

  limbo = false;
  has_nv_state = true;

  status_messages_badly_needed = false;
  not_deprioritize_status_before_this = 0;

  nbreqs = 0;
  nbrounds = 0;
  
  // Read view change, status, and recovery timeouts from replica's portion
  // of "config_file"
  int vt, st, rt;
  fscanf(config_file, "%d\n", &vt);  
  fscanf(config_file, "%d\n", &st);
  fscanf(config_file, "%d\n", &rt);

  // Create timers and randomize times to avoid collisions.
  srand48(getpid());

  delay_pre_prepare_timer = new ITimer(byz_pre_prepare_delay , delay_pre_prepare_timer_handler);

#ifdef DELAY_ADAPTIVE 
//  call_pre_prepare_timer = new ITimer(call_pre_prepare_timer_duration, call_pre_prepare_timer_handler);
  pre_prepare_timer = new ITimer(pre_prepare_timer_duration, pre_prepare_timer_handler);
  pre_prepare_timer->start();
#endif

#ifdef THROUGHPUT_ADAPTIVE
  throughput_timer = new ITimer(throughput_timer_duration, throughput_timer_handler);
  increment_timer = new ITimer(increment_timer_duration, increment_timer_handler);
#endif

  vtimer = new ITimer(vt+lrand48()%100, vtimer_handler);

  stimer = new ITimer(st+lrand48()%100, stimer_handler);

//  stimer->start(); //not started in oldbft, so not starting here


  // Skew recoveries. It is important for nodes to recover in the reverse order
  // of their node ids to avoid a view-change every recovery which would degrade
  // performance.
  rec_ready = false;
  ntimer = new ITimer(30000/max_out, ntimer_handler);

  recovering = false;
  rr = 0;
  rr_views = new View[num_replicas];
  recovery_point = Seqno_max;
  max_rec_n = 0;

  exec_command = 0;
  non_det_choices = 0;

  print_stuff = true;

  // rewind config_file, and read again all the replica lines

  char* fake_string;
  fake_string= (char*)malloc(2000*sizeof(char));
  int fake_int;

  rewind(config_file);

  //now, skip the useless stuff at the beginning
  fscanf(config_file, "%256s\n", fake_string);
  fscanf(config_file, "%d\n", &fake_int);
  fscanf(config_file, "%d\n", &fake_int);
  fscanf(config_file, "%d\n", &fake_int);
  fscanf(config_file, "%256s %d\n", fake_string, &fake_int);


  //ok, now i am in the right place.
  for(int j=0; j< num_replicas; j++){
    int base = j*num_replicas;

    fscanf(config_file, "%256s %32s %d ",fake_string, fake_string, &fake_int);

    for( int k =0; k<num_replicas; k++){
      fscanf(config_file,"%32s %d ",addresses[base+k],&ports[base+k]);
    }

    fscanf(config_file,"%d %d %1024s \n",&fake_int, &fake_int, fake_string);
  }

  free(fake_string); 

  int error;
  excluded_clients = false;

  // initialize the array of sockets used for replica - replica communicaztion
  sock_array = (int*)malloc(num_replicas*sizeof(int));
  excluded_replicas = (bool*)malloc(num_replicas*sizeof(bool));

  int base= id()*4;
  
  printf("I am replica %d\n",id());

  for (int j = 0 ; j < num_replicas ; j++ ) {

    // initializing
    sock_array[j] = socket(AF_INET, SOCK_DGRAM, 0);
    
    Addr tmp;
    tmp.sin_family = AF_INET;
    tmp.sin_port = htons(ports[base+j]); //in any case
    tmp.sin_addr.s_addr = inet_addr(addresses[base+j]);
//    printf("binding on address %s and port %d for communications with replica %d\n",addresses[base+j],ports[base+j],j);
//    printf("packets to replica %d will be sent to address %s and port %d\n",j,addresses[j*4+id()], ports[j*4 + id()]);

    //binding
    int error = bind(sock_array[j], (struct sockaddr*)&tmp, sizeof(Addr));
    if (error < 0) {
      perror("Unable to name socket");
      exit(1);
    }

    //making the socket async
    error = fcntl(sock_array[j], F_SETFL, O_NONBLOCK);
    if (error < 0) {
      perror("unable to set socket to asynchronous mode");
      exit(1);
    }

    int snd_buf_size = 4000000;
    error = setsockopt(sock_array[j], SOL_SOCKET, SO_SNDBUF,
                        (char*)&snd_buf_size, sizeof(snd_buf_size));
    if (error < 0) {
      perror("unable to increase send buffer size");
      exit(1);
    }

      //int rcv_buf_size = 262144;
    int rcv_buf_size = 4000000;
    error = setsockopt(sock_array[j], SOL_SOCKET, SO_RCVBUF,
                        (char*)&rcv_buf_size, sizeof(rcv_buf_size));
    if (error < 0) {
      perror("unable to increase send buffer size");
      exit(1);
    }

    //resetting the excluded_replicas array
    excluded_replicas[j] = false;
  } 

  int snd_buf_size = 4000000;
  error = setsockopt(sock, SOL_SOCKET, SO_SNDBUF,
                      (char*)&snd_buf_size, sizeof(snd_buf_size));
  if (error < 0) {
    perror("unable to increase send buffer size");
    exit(1);
  }

  int rcv_buf_size = 4000000;
  error = setsockopt(sock, SOL_SOCKET, SO_RCVBUF,
                      (char*)&rcv_buf_size, sizeof(rcv_buf_size));
  if (error < 0) {
    perror("unable to increase send buffer size");
    exit(1);
  }
 
  status_pending = (Message**)malloc( num_replicas * sizeof(Message*));
  for (int index=0; index < num_replicas ; index++)
    status_pending[index] = NULL;
  status_to_process = 0;
  s_identity=-1;


  circular_buffer_magic = (Message*)0x12344321; //I don't want to read that value

  circular_buffer = (Message**)malloc(circular_buffer_size*sizeof(Message*));
  for (int i=0; i<circular_buffer_size; i++)
    circular_buffer[i] = circular_buffer_magic;
  cb_write_index = 0;
  cb_read_index = 0;

  printf("[byz_pre_prepare_delay : %ld]\n",byz_pre_prepare_delay);

} 

void Replica::register_exec(int (*e)(Byz_req *, Byz_rep *, Byz_buffer *, int, bool, long int)) { //last long int is the execute delay
  exec_command = e; 
}


void Replica::register_nondet_choices(void (*n)(Seqno, Byz_buffer *), int max_len) {

  non_det_choices = n;
  max_nondet_choice_len = max_len;
}

void Replica::compute_non_det(Seqno s, char *b, int *b_len) {
  if (non_det_choices == 0) {
    *b_len = 0;
    return;
  }
  Byz_buffer buf;
  buf.contents = b;
  buf.size = *b_len;
  non_det_choices(s, &buf);
  *b_len = buf.size;
}


Replica::~Replica() {}
 
void Replica::recv() {
  // Compute session keys and send initial new-key message.  
  Node::send_new_key();

  // Compute digest of initial state and first checkpoint.
  state.compute_full_digest();

  // Start status and authentication freshness timers
  stimer->start();
  atimer->start();
  if (id() == primary())
    ntimer->start();

  // Allow recoveries
  rec_ready = true;

  fprintf(stderr, "Replica ready\n");

  int count_idle=0;
  //int max_queue_size=0;
  //int silly_counter=1;

#ifdef REPLICA_FLOOD_PROTECTION
  int rmcount[5];
  for (int k=0; k<5; k++)
    rmcount[k]=1;
  flood_protection_active=true;
  flood_protection_view=0; 
#endif
  long int client_count = 0;

  Message* m;
  while (1) {
    // [attacks] Check whether we need to perform a specific type of attack.
    if (floodMax()) {
    //if (true) {
      char *buffer = (char*)malloc(sizeof(char)*Max_message_size);
      ((Message_rep*)buffer)->tag=255;
      // flood with max-sized messages
      //Message m(Max_message_size);

      for (int i=0; i<num_replicas; i++) {
        Addr *to =  (Addr *)principals[i]->address();
        to->sin_port = htons(ports[i*4+id()]); //in any case
        to->sin_addr.s_addr = inet_addr(addresses[i*4 + id()]); //change addres because multiple nics are enabled
        sendto(sock_array[i], buffer, Max_message_size,0,(struct sockaddr*)to, sizeof(Addr));
      }
      free(buffer);
      continue;
    }
    else if (floodProtocol()) {
      // flood with protocol messages (status and request)
      send_status(true);
      state.send_fetch(true);
      continue;
    }

    FD_ZERO(&file_descriptors); //initialize file descriptor set

    if (!excluded_clients)
      FD_SET(sock, &file_descriptors); //add sock to the set of file descriptors to listen to

    for (int j = 0; j < num_replicas ; j++)
      if (!excluded_replicas[j])
        FD_SET(sock_array[j], &file_descriptors);

    listen_time.tv_sec = 0;
    listen_time.tv_usec = 50;

    int maxsock = sock;
    for (int j = 0 ; j < num_replicas ; j++)
      if(!excluded_replicas[j])
        maxsock = MAX(maxsock, sock_array[j]);  

    bool idle = true;
    select(maxsock+1, &file_descriptors, NULL, NULL, &listen_time);

    ITimer::handle_timeouts();



    if(!excluded_clients) {
    /*
      if(FD_ISSET(sock, &file_descriptors)) {
        idle = false;
        // something to read in the client socket
        Message* m = new Message(Max_message_size);

        int ret = recvfrom(sock, m->contents(), m->msize(), 0, 0, 0);
        if (ret >= (int)sizeof(Message_rep) && ret >= m->size()) {
          client_count++;
          jump_table(m);
          if( primary() == id()  && has_new_view() && (client_count < np()-n())
	     && (!rqueue.full_batch() && rqueue.size() < (np()-n())/2))
            continue;
        }
        else
          printf("++++++++ WTF? +++++++++ \n");
      }
    */
 
      m = cb_read_message();
      th_assert(m != circular_buffer_magic, "Ooooops... messed up with circular buffer...");
      if(m){
        //buffer was not empty, read something
        client_count++;
        jump_table(m);
         //if( primary() == id()  && has_new_view() && (client_count < np()-n()) && (!rqueue.full_batch() && rqueue.size() < (np()-n())/2))
         if( primary() == id()  && has_new_view() && !rqueue.full_batch() )
           continue;
      }


    }
   
    for(int j=0; j < num_replicas ; j++) {
      // for each replica
      if(!excluded_replicas[j]){
        //I want to listen at this replica
        if(FD_ISSET(sock_array[j], &file_descriptors)) {
          client_count = 0; 
          idle = false;
#ifdef REPLICA_FLOOD_PROTECTION
          if (flood_protection_active) { 
            rmcount[j] ++;
            rmcount[4] ++;
            if(rmcount[4] % check_rmcount == 0) {
              // printf("%d %d %d %d %d\n", rmcount[0], rmcount[1], rmcount[2], rmcount[3], rmcount[4]);
              for (int k=0; k<4 && k!=id() ; k++){
                if (rmcount[k] > ((float)rmcount[4] * rmcount_threshold) ){
                  printf(" ********* FLOOD PROTECTION: excluding replica %d\n",k);
                  excluded_replicas[k]=true;
                  flood_protection_active=false;
                  flood_protection_view= view(); 
                  //system("./mute_nic eth5 &");
                }
              }
              for (int j=0;j<5;j++) 
                rmcount[j]=1;
            }
          }
#endif
          //printf("Received something from replica %d\n",j);
          // something to read from replica j
          
          if (!excluded_replicas[j]) {
            Message* mp = new Message(Max_message_size);
            int ret = recvfrom(sock_array[j], mp->contents(), mp->msize(), 0, 0, 0);
            if (ret >= (int)sizeof(Message_rep) && ret >= mp->size()) {
              jump_table(mp);
            }
            else {
             delete mp; 
             // printf("--------- WTF? --------- ret: %d from replica %d\n",ret,j);
              if (ret < 0)
                printf("errno: %s\n",strerror(errno));
            }
          }
        }
      }
    }


    //checked clients and all the replicas, if idle is still true, I have nothing to do...
    if (idle)
      count_idle++;
    else
      count_idle = 0;
    if (count_idle > 3) {
      Message* m = pick_next_status();
      if (m) {
        gen_handle<Status>(m);
      }
    }
  }
}

void Replica::send(Message *m, int i) {
  th_assert(i == All_replicas || (i >= 0 && i < num_principals),
            "Invalid argument");
  m->check_msg();
  
  if (i == All_replicas) {
    for (int x=0; x<num_replicas; x++){
    //  if (x != id()) //some code relies on receiving self-created messages
       replica->send(m,x);
    }
    return;
  }

  th_assert(i != All_replicas,"the multisocket version does not work well with multicast... \n");
  Addr *to =  (Addr *)principals[i]->address();
//    (i == All_replicas) ? (Addr *)group->address() : (Addr *)principals[i]->address();

  int error = 0;
  int size = m->size();
  //if(m->tag() == View_change_tag) printf("node sending view change of size %d to %d\n",size,i);
  while (error < size) {
    //INCR_OP(num_sendto);
    //INCR_CNT(bytes_out,size);
    //START_CC(sendto_cycles);
    if (i < num_replicas) { //sending to a replica
      to->sin_port = htons(ports[i*4+id()]); //in any case
      to->sin_addr.s_addr = inet_addr(addresses[i*4 + id()]); //change addres because multiple nics are enabled
      error = sendto(sock_array[i], m->contents(), size,0,(struct sockaddr*)to, sizeof(Addr));
      //printf("sent %d bytes to replica %d\n",error,i);
    }
    else  //sending to a client
      error = sendto(verifier->client_socket, m->contents(), size, 0, (struct sockaddr*)to, sizeof(Addr));

    if (error == EAGAIN)
      printf("need to send this again...\n");
    if (error < 0 && error != EAGAIN)
    //STOP_CC(sendto_cycles);
//#ifndef NDEBUG
    if (error < 0 && error != EAGAIN)
      perror("Node::send: sendto");
//#endif
  }
}


void Replica::jump_table(Message* m) {

  th_assert(m,"null message...");
   
  // TODO: This should probably be a jump table.
  switch (m->tag()) {

  case Request_tag:
    //printf("A");
    gen_handle<Request>(m);
    break;

  case Pre_prepare_tag:
    //printf("B");
    gen_handle<Pre_prepare>(m);
    break;
      
  case Prepare_tag:
    //printf("C");
    gen_handle<Prepare>(m);
    break;
      
  case Commit_tag:
    //printf("D");
    gen_handle<Commit>(m);
    break;

  case Checkpoint_tag:
    //printf("E");
    gen_handle<Checkpoint>(m);
    break;

  case New_key_tag:
    //printf("F");
    gen_handle<New_key>(m);
    break;

  case View_change_ack_tag:
    //printf("G");
    gen_handle<View_change_ack>(m);
    break;

  case Status_tag:
    //printf("H");

    if (status_messages_badly_needed){
      gen_handle<Status>(m);
    } else  {
      //for now, just add this message to status_pending
      s_identity = ((Status*)m)->id();
      th_assert(s_identity >=0 && s_identity < num_replicas, "meaningless s_identity\n");
      if (status_pending[s_identity] != NULL) {
        delete status_pending[s_identity];
        //printf("removed older status from %d to status_pending\n",identity);
      }
      //printf("added status from %d to status_pending\n",identity);
      status_pending[s_identity]=m;
    }
    break;

  case Fetch_tag:
    //printf("I");

    gen_handle<Fetch>(m);
    break;

  case Reply_tag:
    //printf("L");
    gen_handle<Reply>(m);
    break;

  case Meta_data_tag:
    //printf("O");
    gen_handle<Meta_data>(m);
    break;

  case Meta_data_d_tag:
    //printf("P");
    gen_handle<Meta_data_d>(m);
    break;

  case Data_tag:
    //printf("Q");
    gen_handle<Data>(m);
    break;

  case View_change_tag:
    //printf("R");
    gen_handle<View_change>(m);
    break;

  case New_view_tag:
    //printf("S");
    gen_handle<New_view>(m);
    break;

  case Wrapped_request_tag:
    //printf("**** Main Thread should not receive wrapped requests...");
    delete m;
    break;

  default:
    // Unknown message type.
    //printf("**** Received garbage, deleting\n");
    delete m;
  }
}

void Replica::handle(Request *m) {

  //if (print_stuff)
//    printf("\t\t\t\t\t\thandling request\n");

  int cid = m->client_id();
  bool ro = m->is_read_only();
  Request_id rid = m->request_id();

  // [attacks] Bias on clients?
  if (clientBias(m)) return;

  if (has_new_view() ){

    // Replica's requests must be signed and cannot be read-only.  
    if (!is_replica(cid) || (m->is_signed() & !ro)) {
#ifdef SPECULATIVE
#else
      if (ro) {
	// Read-only requests.
	if (execute_read_only(m) || !ro_rqueue.append(m))
 	  delete m;

 	return;
      }
#endif
        Request_id last_rid = replies.req_id(cid);
        if (last_rid < rid) {
 	// Request has not been executed.   
 	if (id() == primary()) {
 	  if (!rqueue.in_progress(cid, rid, v) && rqueue.append(m)) {
 	    //	    fprintf(stderr, "RID %qd. ", rid);
 	    send_pre_prepare();
 	    return;
 	  }
 	} else {
 	  if (m->size() > Request::big_req_thresh	&& brt.add_request(m))
 	    return;

 	  if (rqueue.append(m)) {
 	    if (!limbo) {

  	      Req_queue tmp; //req_queue used just to store the request, because wrapped request constructor takes a request queue as a parameter
              		     //It can be easily modified to accept a request* instead of a request queue, but I prefer to do exactly the same things
               	   	     //that are done for the generation of a Pre_prepare message.
  	      Request *cloned = m->clone();
  	      tmp.append(cloned);
              Wrapped_request* wrapped = new Wrapped_request((Seqno)id(),(View)1,tmp);
 	      ((Node*)node)->send(wrapped, primary());
              delete wrapped;

 	      //vtimer->start(); //??? necessary? useful?
 	    }
 	    return;
 	  }
 	}
       } else if (last_rid == rid) {
 	// Retransmit reply.
 	replies.send_reply(cid, view(), id());

 	if (id() != primary() && !replies.is_committed(cid)	&& rqueue.append(m)) {
 	  //printf("vtimer restart (5)\n");
 	  vtimer->start();
 	  return;
 	}
       }
     }
   } else {
     if (m->size() > Request::big_req_thresh && !ro && brt.add_request(m, false)){
       th_assert(false, "big requests are disabled");
       return;
     }
   }

   delete m;
}



void Replica::send_pre_prepare(bool force_send) {
  if ( !has_new_view() ) {
//    printf("Should not be here!!!!!\n");
    return;
  }
  th_assert(primary() == id(), "Non-primary called send_pre_prepare");

  // if the parameter byz_pre_prepare_delay is not 0, this
  // is a byzantine replica that waits a delay of
  // byz_pre_prepare_delay us before sending the pre_prepare
  if (byz_pre_prepare_delay && !force_send){
    delay_pre_prepare_timer->restart();
    return;
  }

#ifdef SPECULATIVE
  if ((force_send || rqueue.size() > 0) && seqno+1 <= last_committed+congestion_window &&
      seqno+1 <= max_out+last_stable && has_new_view()) {
#else
  // If rqueue is empty there are no requests for which to send
  // pre_prepare and a pre-prepare cannot be sent if the seqno excedes
  // the maximum window or the replica does not have the new view.
    if ((force_send || rqueue.size() > 0) && seqno+1 <= last_executed+congestion_window && 
      seqno+1 <= max_out+last_stable && has_new_view()) {
#endif
    //  printf("requeu.size = %d\n", rqueue.size());
    //  if (seqno % checkpoint_interval == 0)
    //if (print_stuff)
    //  printf("SND: PRE-PREPARE seqno=%qd last_stable=%qd\n", seqno+1, last_stable);

	long old_size = rqueue.size();

	if (nbreqs >= 10000) {
		printf("Avg batch sz: %f\n", (float)nbreqs/(float)nbrounds);
		nbreqs = nbrounds = 0;
	}
	

	// Create new pre_prepare message for set of requests
    // in rqueue, log message and multicast the pre_prepare.
    seqno++;
    //    fprintf(stderr, "Sending PP seqno %qd\n", seqno);

#ifdef SPECULATIVE
    Pre_prepare *pp = new Pre_prepare(view(), seqno, rqueue, replies.history_digest());
    if (seqno % checkpoint_interval == 0){
      Digest h;
      h.set(seqno);
      replies.set_history(h);
    }else
      replies.append_digest(replies.history_digest(),
			    pp->digest());
   
#else
    Pre_prepare *pp = new Pre_prepare(view(), seqno, rqueue);
#endif
    nbreqs += old_size - rqueue.size();
    nbrounds++;
    // TODO: should make code match my proof with request removed 
    // only when executed rather than removing them from rqueue when the
    // pre-prepare is constructed.

    //if (print_stuff)
    //    printf("SND:  pp: (%qd, %qd),  last stable: %qd\n",seqno, pp->view(),  last_stable);
#ifdef SPECULATIVE
    //  printf("  :   history "); replies.history_digest().print(); printf("\n");
#endif
    send(pp, All_replicas);
    plog.fetch(seqno).add_mine(pp);
//    vtimer->stop(); //not in old pbft, so commented here
  }
}


template <class T> 
bool Replica::in_w(T *m) {
  const Seqno offset = m->seqno() - last_stable;

  if (offset > 0 && offset <= max_out)
    return true;

  if (offset > max_out && m->verify()) {
    // Send status message to obtain missing messages. This works as a
    // negative ack.
    send_status();
  }
  
  return false;
}


template <class T> 
bool Replica::in_wv(T *m) {
  const Seqno offset = m->seqno() - last_stable;

  if (offset > 0 && offset <= max_out && m->view() == view())
    return true;

  if ((m->view() > view() || offset > max_out) && m->verify()) {
    // Send status message to obtain missing messages. This works as a
    // negative ack.
    send_status();
  }

  return false;
}


void Replica::handle(Pre_prepare *m) {
  const Seqno ms = m->seqno();
  
  //if(print_stuff)
  //  printf("\tRCV: pp (%qd, %qd) from %d\n",ms,view(),m->id());
  
  Byz_buffer b;
  
  b.contents = m->choices(b.size);
  
#ifdef SPECULATIVE
  if (in_wv(m) && ms > low_bound && has_new_view() && ms <= seqno+1){
    Prepared_cert& pc = plog.fetch(ms);
    

    // Accept a message if it is from the past and might finish of a view change
    // or if it is the next message in sequence
    if (ms < seqno +1 ||
	(ms == seqno +1 && replies.history_digest() == m->history_digest())){

      if (pc.add(m)) {
#ifdef DELAY_ADAPTIVE
	received_pre_prepares++;
#endif
	//      printf("\t\t\t\tsending prepare\n");                   
	send_prepare(pc);
	
	if (pc.is_complete())
	  send_commit(ms);
	if (ms == seqno+1){
	  seqno++;
	  // if ms is a checkpoint, then the history is special
	  if (ms % checkpoint_interval == 0){
	    Digest h;
	    h.set(ms);
	    replies.set_history(h);
	  }else
	    // otherwise the history is message history + message digest
	    replies.append_digest(m->history_digest(), m->digest());
	}
      }
      return;
    }
  }



#else // =================================================================
    if (in_wv(m) && ms > low_bound && has_new_view()) {
      Prepared_cert& pc = plog.fetch(ms);

    // Only accept message if we never accepted another pre-prepare
    // for the same view and sequence number and the message is valid.
    if (pc.add(m)) {
#ifdef DELAY_ADAPTIVE
      received_pre_prepares++;
#endif
      //      printf("\t\t\t\tsending prepare\n");
      send_prepare(pc);
      
      if (pc.is_complete())
	send_commit(ms);
    }
    return;
  }
#endif //==============================================================

  if (!has_new_view()) {
    // This may be an old pre-prepare that replica needs to complete
    // a view-change.
    vi.add_missing(m);
    return;
  }
  
  delete m;
}


void Replica::send_prepare(Prepared_cert& pc) {
  if (pc.my_prepare() == 0 && pc.is_pp_complete()) {
    // Send prepare to all replicas and log it.
    Pre_prepare* pp = pc.pre_prepare();

#ifdef SPECULATIVE
    Prepare *p = new Prepare(v, pp->seqno(), pp->digest(), pp->history_digest());
#else
    Prepare *p = new Prepare(v, pp->seqno(), pp->digest());
#endif
    //if (print_stuff)
    //  printf("\tSND prepare (%qd, %qd)\n", p->seqno(), p->view());

    send(p, All_replicas);
    pc.add_mine(p);
  }
}


void Replica::send_commit(Seqno s) {

#ifdef SPECULATIVE
#else
  // Executing request before sending commit improves performance
  // for null requests. May not be true in general.
  if (s == last_executed+1)
    execute_prepared(false);
#endif

  Commit* c = new Commit(view(), s);
  //if (print_stuff)
  //  printf("\t\tSND commit (%qd, %qd)\n", s, c->view());
  send(c, All_replicas);
  
  if (s > last_prepared)
    last_prepared = s;

  Certificate<Commit>& cs = clog.fetch(s);
  if (cs.add_mine(c) && cs.is_complete()) {
    execute_committed();
  }
}
 

void Replica::handle(Prepare *m) {
  const Seqno ms = m->seqno();
  //if (print_stuff)
  //  printf("\t\tRCV p (%qd, %qd) from %d\n", ms, m->view(), m->id());

  // Only accept prepare messages that are not sent by the primary for
  // current view.
  if (in_wv(m) && ms > low_bound && primary() != m->id() && has_new_view()) {
    Prepared_cert& ps = plog.fetch(ms);
    if (ps.add(m) && ps.is_complete()) 
      send_commit(ms);
    return;
  }

  if (m->is_proof() && !has_new_view()) {
    // This may be an prepare sent to prove the authenticity of a
    // request to complete a view-change.
    vi.add_missing(m);
    return;
  }

  delete m;
  return;
}


void Replica::handle(Commit *m) {
  const Seqno ms = m->seqno();
  
  //if(print_stuff)
  //  printf("\t\t\tRCV c (%qd, %qd) from %d\n", ms, m->view(), m->id());

  // Only accept messages with the current view.  TODO: change to
  // accept commits from older views as in proof.
  //Seqno commit_seqno = m->seqno();
  if (in_wv(m) && ms > low_bound) {   
    Certificate<Commit>& cs = clog.fetch(m->seqno());
    if (cs.add(m) && cs.is_complete()) {
      execute_committed();
    }
    return;
  }
  delete m;
  return;
}


void Replica::handle(Checkpoint *m) {
  const Seqno ms = m->seqno();
  //if (print_stuff)
  //  printf("\t\t\t\tRCV CP %qd from %d.  last_stable: %qd, last_exec: %qd\n",ms, m->id(), last_stable, last_executed);
  if (ms > last_stable) {
    //printf("CHECKPOINT 1\n");
    if (ms <= last_stable + max_out) {
      //printf("CHECKPOINT 2\n");
      // Checkpoint is within my window.  Check if checkpoint is
      // stable and it is above my last_executed.  This may signal
      // that messages I missed were garbage collected and I should
      // fetch the state
#ifdef SPECULATIVE
      bool late = m->stable() && last_committed < ms;
#else
      bool late = m->stable() && last_executed < ms;
#endif

      //printf("CHECKPOINT 3a\n");
      if (!late) {
	//printf("CHEcKPOINT 4\n");
	Certificate<Checkpoint> &cs = elog.fetch(ms);
	if (cs.add(m) && cs.mine() && cs.is_complete()) {
	  //printf("CHECKPOINT 5\n");
	  // I have enough Checkpoint messages for m->seqno() to make it stable.
	  // Truncate logs, discard older stable state versions.
	  //	  fprintf(stderr, "CP MSG call MS %qd!!!\n", last_executed);
	  mark_stable(ms, true);
	}
	return;
      }
    }
    
    if (m->verify()) {
      //printf("CHECKPOINT 6\n");
      // Checkpoint message above my window.

      if (!m->stable()) {
	//printf("CHECKPOINT 7\n");
	// Send status message to obtain missing messages. This works as a
	// negative ack.
	send_status();
	delete m;
	return;
      } 

      // Stable checkpoint message above my last_executed.
      Checkpoint *c = sset.fetch(m->id());
      if (c == 0 || c->seqno() < ms) {
	//printf("CHECKPOINT 8\n");
	delete sset.remove(m->id());
	sset.store(m);
	if (sset.size() > f()) {
	  //printf("CHECKPOINT 9\n");
#ifdef SPECULATIVE
	  if (last_tentative_execute > last_committed){
#else//}
	  if (last_tentative_execute > last_executed) {
#endif
	    // Rollback to last checkpoint
	    //printf("CHECKPOINT 10\n");
	    th_assert(!state.in_fetch_state(), "Invalid state");
	    Seqno rc = state.rollback();
#ifdef SPECULATIVE
	    last_tentative_execute = last_committed = rc;
#else
	    last_tentative_execute = last_executed = rc;
#endif
	    //	    fprintf(stderr, ":):):):):):):):) Set le = %d\n", last_executed);
	  }
	  //printf("CHECKPOINT 10a\n");
	  // Stop view change timer while fetching state. It is restarted
          // in new state when the fetch ends.
	  vtimer->stop(); 
#ifdef SPECULATIVE
	  //1
	  if (!has_nv_state || !has_new_view()){
	    printf("preforming a state fetch has_nv_state %d has_new_view %d\n",
		   has_nv_state, has_new_view());
	    state.start_fetch(last_committed);
	  }else{
	    printf("starting a state fetch while i have the new view state\n");
	    send_view_change();
	  }
#else
	  state.start_fetch(last_executed);
#endif
	}
	//printf("checkpoint 11\n");
	return;
      }
    }
  }
  delete m;
  return;
}


void Replica::handle(New_key *m) {
  //printf("\t\t\t\t\t\tnew key\n");
  if (!m->verify()) {
    //printf("BAD NKEY from %d\n", m->id());
  } 
  delete m;
}


void Replica::handle(Status* m) {
  //printf("\t\t\t\t\t\tstatus from %d\n", m->id());
  static const int max_ret_bytes=65536; 

  if (m->verify()) {
    Time current;
    Time *t;
    current = currentTime();
    Principal *p = node->i_to_p(m->id());

    // Retransmit messages that the sender is missing. 
    if (last_stable > m->last_stable() + max_out) {
      // Node is so out-of-date that it will not accept any
      // pre-prepare/prepare/commmit messages in my log.
      // Send a stable checkpoint message for my stable checkpoint.
      Checkpoint *c = elog.fetch(last_stable).mine(&t);
      //th_assert(c != 0 && c->stable(), "Invalid state");
      if( c != 0 && c->stable() ) // randomly triggered assertion replaced with this line...
        retransmit(c, current, t, p);
      delete m;
      return; 
    }

    // Retransmit any checkpoints that the sender may be missing.
    int max = MIN(last_stable, m->last_stable())+max_out;
    int min = MAX(last_stable, m->last_stable()+1);
    for (Seqno n=min; n <= max; n++) {
      if (n%checkpoint_interval == 0) {
	Checkpoint *c = elog.fetch(n).mine(&t);
	if (c != 0) {
	  retransmit(c, current, t, p);
	  th_assert(n == last_stable || !c->stable(), "Invalid state");
	}
      }
    }

    if (m->view() < v) {
      // Retransmit my latest view-change message
      View_change* vc = vi.my_view_change(&t);
      if (vc != 0) 
	retransmit(vc, current, t, p);
      delete m;
      return;
    }

    if (m->view() == v) {
      if (m->has_nv_info()) {
	min = MAX(last_stable+1, m->last_executed()+1);
	for (Seqno n=min; n <= max; n++) {
	  if (m->is_committed(n)) {
	    // No need for retransmission of commit or pre-prepare/prepare
	    // message.
	    continue;
	  }
	  
	  Commit *c = clog.fetch(n).mine(&t);
	  if (c != 0) {
	    retransmit(c, current, t, p);
	  }
	  
	  if (m->is_prepared(n)) {
	    // No need for retransmission of pre-prepare/prepare message.
	    continue;
	  }
	  
	  // If I have a pre-prepare/prepare send it, provided I have sent
	  // a pre-prepare/prepare for view v.
	  if (primary() == node_id) {
	    Pre_prepare *pp = plog.fetch(n).my_pre_prepare(&t);
	    if (pp != 0) {
	      retransmit(pp, current, t, p);
	    }
	  } else {
	    Prepare *pr = plog.fetch(n).my_prepare(&t);
	    if (pr != 0) {
	      retransmit(pr, current, t, p);
	    }
	  }
	}

	if (id() == primary()) {
	  // For now only primary retransmits big requests.
	  Status::BRS_iter gen(m);
	  
	  int count = 0;
	  Seqno ppn;
	  BR_map mrmap;
	  while (gen.get(ppn, mrmap) && count <= max_ret_bytes) {
	    if (plog.within_range(ppn)) {
	      Pre_prepare_info::BRS_iter gen(plog.fetch(ppn).prep_info(), mrmap);
	      Request* r;
	      while (gen.get(r)) {
		send(r, m->id());
		count += r->size();
	      }
	    }
	  }
	}
      } else {
	// m->has_nv_info() == false
	if (!m->has_vc(node_id)) {
	  // p does not have my view-change: send it.
	  View_change* vc = vi.my_view_change(&t);
	  //View_change* vc = vi.my_view_change();
	  th_assert(vc != 0, "Invalid state");
	  retransmit(vc, current, t, p);
	}

	if (!m->has_nv_m()) {
	  if (primary(v) == node_id && vi.has_new_view(v)) {
	    // p does not have new-view message and I am primary: send it
	    New_view* nv = vi.my_new_view(&t);
	    //New_view* nv = vi.my_new_view();
	    if (nv != 0) 
	      retransmit(nv, current, t, p);
	  }
	} else {
	  if (primary(v) == node_id && vi.has_new_view(v)) {
	    // Send any view-change messages that p may be missing
	    // that are referred to by the new-view message.  This may
	    // be important if the sender of the original message is
	    // faulty.


	  } else {
	    // Send any view-change acks p may be missing. 
	    for (int i=0; i < num_replicas; i++) {
	      if (m->id() == i) continue;
	      View_change_ack* vca = vi.my_vc_ack(i);
	      if (vca && !m->has_vc(i)) 
		// View-change acks are not being authenticated
		retransmit(vca, current, &current, p);
	    }
	  }

	  // Send any pre-prepares that p may be missing and any proofs
          // of authenticity for associated requests.  
	  Status::PPS_iter gen(m);

	  int count = 0;
	  Seqno ppn;
	  View ppv;
	  bool ppp;
	  BR_map mrmap;
	  while (gen.get(ppv, ppn, mrmap, ppp)) {
	    Pre_prepare* pp = 0;
	    if (m->id() == primary(v))
	      pp = vi.pre_prepare(ppn, ppv);
	    else {
	      if (primary(v) == id() && plog.within_range(ppn))
		pp = plog.fetch(ppn).pre_prepare();
	    }
	      
	    if (pp) { 
	      retransmit(pp, current, &current, p);

	      if (count < max_ret_bytes && mrmap != ~0) {
                 Pre_prepare_info pi;
                 pi.add_complete(pp);
 
                 Pre_prepare_info::BRS_iter gen(&pi, mrmap);
                 Request* r;
                 while (gen.get(r)) {
                   send(r, m->id());
		   count += r->size();
		 }
		 pi.zero(); // Make sure pp does not get deallocated
	      }
	    }

	    if (ppp) 
	      vi.send_proofs(ppn, ppv, m->id());	      
	  }
	}
      }
    }
  } else {
    // It is possible that we could not verify message because the
    // sender did not receive my last new_key message. It is also
    // possible message is bogus. We choose to retransmit last new_key
    // message.  TODO: should impose a limit on the frequency at which
    // we are willing to do it to prevent a denial of service attack.
    // This is not being done right now.
    if (last_new_key != 0 && !m->verify()) {
     	//printf("sending new key to %d\n", m->id());
	send(last_new_key, m->id());
    }
  }

  delete m;
}


void Replica::handle(View_change *m) {

  //muting replicas

  if(has_new_view() && m->view() > view()){
    excluded_replicas[m->id()] = true;
    printf("muted replica %d for vie change %qd \n",m->id(), m->view());
  }

  printf("RECV: view change v=%qd from %d\n", m->view(), m->id());
  bool modified = vi.add(m);
  if (!modified)
    return;
    
  // TODO: memoize maxv and avoid this computation if it cannot change i.e.
  // m->view() <= last maxv. This also holds for the next check.
  View maxv = vi.max_view();
  if (maxv > v || m->id() == primary())  {
    // Replica has at least f+1 view-changes with a view number
    // greater than or equal to maxv: change to view maxv.
    // or the current primary is asking for a view change
    v = maxv-1;
    vc_recovering = true;
    //printf("joining a view change\n");
    send_view_change();

    return;
  }
 
  if (limbo && primary() != node_id) {
    maxv = vi.max_maj_view();
    th_assert(maxv <= v, "Invalid state");

    if (maxv == v) {
      // Replica now has at least 2f+1 view-change messages with view  greater than
      // or equal to "v"

      // Start timer to ensure we move to another view if we do not
      // receive the new-view message for "v".
      //printf("starting vtimer 1\n");
      //printf("vtimer restart (6)\n");
      vtimer->restart();
      limbo = false;
      vc_recovering = true;
    }
  }
}


void Replica::handle(New_view *m) {
  //printf("\t\t\t\t\t\tnew view\n");
//  printf("RECV: new view v=%qd from %d\n", m->view(), m->id());

  vi.add(m);

#ifdef FAIRNESS_ADAPTIVE

  for (int j=0; j < num_principals; j++){
    //I may have stale informations in my execute_snapshot array... resetting it.
    executed_snapshot[j] = -1;
  }
  
#endif
}


void Replica::handle(View_change_ack *m) {
  //printf("RECV: view-change ack v=%qd from %d for %d\n", m->view(), m->id(), m->vc_id());

  vi.add(m);
}


void Replica::send_view_change() {
  // Move to next view.

  //unmuting replicas
  for(int j=0; j<num_replicas; j++)
    excluded_replicas[j]=false;
  printf("all replicas unmuted\n");


#ifdef DELAY_ADAPTIVE
  pre_prepare_timer->restop();
#endif
  v++;
#ifdef SPECULATIVE
  printf("sending a view_change, v: %qd, last_executed: %qd, seqno %qd \n",
	 v,last_committed, seqno);
#endif
  status_messages_badly_needed = true; // avoid status message de-prioritization

  //muting clients
  excluded_clients = true;

#ifdef THROUGHPUT_ADAPTIVE
  first_checkpoint_after_view_change = true;
  vc_already_triggered = false;
  throughput_timer->restop();
#endif

  cur_primary = v % num_replicas;
  limbo = true;
  vtimer->stop(); // stop timer if it is still running
  //printf("stopping ntimer (2)\n");
  ntimer->restop();

#ifdef SPECULATIVE
  if (last_tentative_execute > last_committed) {

    th_assert(!state.in_fetch_state(), "Invalid state");
    Seqno rc = state.rollback();
    printf("XXXRolled back in vc to %qd with last_executed=%qd %qd\n", rc, last_committed, seqno);
    last_tentative_execute = last_committed = rc;
  }
  last_prepared = last_committed;
#else
  if (last_tentative_execute > last_executed) {
    // Rollback to last checkpoint
    th_assert(!state.in_fetch_state(), "Invalid state");
    Seqno rc = state.rollback();
    printf("XXXRolled back in vc to %qd with last_executed=%qd %qd\n", rc, last_executed, seqno);
    last_tentative_execute = last_executed = rc;
  }
  last_prepared = last_executed;
#endif


  for (Seqno i=last_stable+1; i <= last_stable+max_out; i++) {
    Prepared_cert &pc = plog.fetch(i);
    Certificate<Commit> &cc = clog.fetch(i);

#ifdef SPECULATIVE
    if (pc.is_complete()) {
      vi.add_complete(pc.rem_pre_prepare());
    }else{
      Prepare *p = pc.my_prepare();
      if (p!= 0){
	vi.add_incomplete(i, p->digest(), p->history_digest());
      }else{
	Pre_prepare *pp = pc.my_pre_prepare();
	if (pp != 0){
	  vi.add_incomplete(i, pp->digest(), pp->history_digest());
	}
      }
    }
#else
    if (pc.is_complete()) {
      vi.add_complete(pc.rem_pre_prepare());
    } else {
      Prepare *p = pc.my_prepare();
      if (p != 0) {
	vi.add_incomplete(i, p->digest());
      } else {
	Pre_prepare *pp = pc.my_pre_prepare();
	if (pp != 0) {
	  vi.add_incomplete(i, pp->digest());
	}
      }
    }
#endif
   
    pc.clear();
    cc.clear(); 
    // TODO: Could remember info about committed requests for efficiency.
  }

  // Create and send view-change message.
  printf("XXX SND: view change %qd seqno %qd\n", v, seqno);
#ifdef SPECULATIVE
  printf("       : history ");replies.history_digest().print(); printf("\n");

  vi.view_change(v, last_committed, &state);
#else
  vi.view_change(v, last_executed, &state);
#endif
}


void Replica::process_new_view(Seqno min, Digest d, Seqno max, Seqno ms) {
  th_assert(ms >= 0 && ms <= min, "Invalid state");
  printf("XXX process new view: %qd, min: %qd, max: %qd, ms: %qd\n", v, min, max, ms);
  not_deprioritize_status_before_this = max + 2*checkpoint_interval ; //just to be sure... ;)

#ifdef REPLICA_FLOOD_PROTECTION
  if(!flood_protection_active && view() > flood_protection_view + flood_protected_views){
    flood_protection_active = true;
    printf(" ******* FLOOD ******* UNMUTING!!! *****\n");
    for (int i=0 ; i < 5 ; i++)
      excluded_replicas[i] = false;
    //system("./unmute_nic eth5 &");
  }
#endif

  vtimer->restop();
  limbo = false;
  vc_recovering = true;

  if (primary(v) == id()) {
    New_view* nv = vi.my_new_view();
    send(nv, All_replicas);
  }

  // Setup variables used by mark_stable before calling it.
  seqno = max-1; 
  if (last_stable > min) 
    min = last_stable;
  low_bound = min;

  if (ms > last_stable) {
    // Call mark_stable to ensure there is space for the pre-prepares
    // and prepares that are inserted in the log below.
#ifdef SPECULATIVE
    mark_stable(ms, last_committed >= ms);
#else
    mark_stable(ms, last_executed >= ms);
#endif
  }

  // Update pre-prepare/prepare logs.
  th_assert(min >= last_stable && max-last_stable-1 <= max_out, "Invalid state");

#ifdef SPECULATIVE
  Digest history;
  if (min % checkpoint_interval == 0)
    history.set(min);
  else
    th_assert(0, "cannot handle view changes with non-standard checkpoints");
  printf("initial history "); history.print(); printf("\n");
#endif

  for (Seqno i=min+1; i < max; i++) {
    Digest d;
#ifdef SPECULATIVE
    Digest h;
    //    printf("base history ");history.print();printf("\n");
    Pre_prepare *pp = vi.fetch_request(i, d, h);
    printf("   pp d %qd ", i);d.print(); printf("\n");
    printf("   pp h %qd ", i);h.print(); printf("\n");
#else
    Pre_prepare* pp = vi.fetch_request(i,d);
#endif
    Prepared_cert& pc = plog.fetch(i);
    
    if (primary() == id()) {
      pc.add_mine(pp);
#ifdef SPECULATIVE
      history.combine(history, d);

    } else {
      
      Prepare *p = new Prepare(v, i, d, h);
      printf("   p d %qd ", i); d.print(); printf("\n");
      printf("   p h %qd ", i);h.print(); printf("\n");
      history.combine(history, d);
#else
    } else {
      Prepare* p = new Prepare(v, i, d);
#endif
      pc.add_mine(p);
      send(p, All_replicas);
      
      th_assert(pp != 0 && pp->digest() == p->digest(), "Invalid state");
      pc.add_old(pp);
    }
  }

#ifdef SPECULATIVE // ====================
  if (last_committed < min){
    has_nv_state = false;
    state.start_fetch(last_committed, min, &d, min <= ms);    
  }else{
    has_nv_state = true;
    if (primary() == id()){
      send_pre_prepare();
      ntimer->start();
    }
    replies.set_history(history);
  }
  
#else  // ================================
  if (primary() == id()) { 
    send_pre_prepare();
    ntimer->start();
  }

  
  if (last_executed < min) {
    has_nv_state = false;
    state.start_fetch(last_executed, min, &d, min <= ms);
  } else {
    has_nv_state = true;

    // Execute any buffered read-only requests
    for (Request *m = ro_rqueue.remove(); m != 0; m = ro_rqueue.remove()) {
      execute_read_only(m);
      delete m;
    }
  }
#endif  // ================================


  if (primary() != id() && rqueue.size() > 0) {
    //printf("vtimer restart (0)\n");
    vtimer->restart();
  }

#ifdef DELAY_ADAPTIVE
  pre_prepare_timer->restop();
  received_pre_prepares = 0;
  if (id() != primary())
    pre_prepare_timer->restart();
    //call_pre_prepare_timer->restart(); 
#endif

#ifdef THROUGHPUT_ADAPTIVE
  vc_already_triggered = false;
  time_to_increment = false;
  increment_timer->restart();
  checkpoints_in_new_view = 0;
#endif

  //unmuting clients
  excluded_clients = false;

 print_stuff = true;

#ifdef SPECULATIVE
 printf("XXX DONE:process new view: %qd, last executed %qd, has new view: %d\n", 
	v, last_committed, has_new_view());

 printf("        :history "); replies.history_digest().print(); printf("\n");
#endif
}


Pre_prepare* Replica::prepared(Seqno n) {
  Prepared_cert& pc = plog.fetch(n);
  if (pc.is_complete()) {
    return pc.pre_prepare();
  }
  return 0;
}


Pre_prepare *Replica::committed(Seqno s) {
  // TODO: This is correct but too conservative: fix to handle case
  // where commit and prepare are not in same view; and to allow
  // commits without prepared requests, i.e., only with the
  // pre-prepare.
  Pre_prepare *pp = prepared(s);
  if (clog.fetch(s).is_complete())
    return pp;
  return 0;
}

 
#ifdef SPECULATIVE
#else
bool Replica::execute_read_only(Request *req) {
	// JC: won't execute read-only if there's a current tentative execution
	// this probably isn't necessary if clients wait for 2f+1 RO responses 
  if (last_tentative_execute == last_executed && 
      !state.in_fetch_state() ) {
    // Create a new Reply message.
    Reply *rep = new Reply(view(), req->request_id(), node_id);
    // Obtain "in" and "out" buffers to call exec_command
    Byz_req inb;
    Byz_rep outb;
    
    inb.contents = req->command(inb.size);	
    outb.contents = rep->store_reply(outb.size);
    
    // Execute command.
    int cid =  req->client_id();
    Principal *cp = i_to_p(cid);
    int error = exec_command(&inb, &outb, 0, cid, true, (long int)0);

    if (outb.size % ALIGNMENT_BYTES)
      for (int i=0; i < ALIGNMENT_BYTES - (outb.size % ALIGNMENT_BYTES); i++)
	outb.contents[outb.size+i] = 0;

    if (!error) {
      // Finish constructing the reply and send it.
      rep->authenticate(cp, outb.size, true);
      if (outb.size < 50 || req->replier() == node_id || req->replier() < 0) {
	// Send full reply.
	send(rep, cid);
      } else {
	// Send empty reply.
	Reply empty(view(), req->request_id(), node_id, 
		    rep->digest(), cp, true);
	send(&empty, cid);
      }
    }

    delete rep;
    return true;
  } else {
    return false;
  }
}
#endif


#ifdef SPECULATIVE
 void Replica::execute_speculative(bool committed){
   if (last_tentative_execute < last_committed + 1 && 
       last_committed < last_stable + max_out && 
       !state.in_fetch_state() && has_new_view()){
     Pre_prepare *pp = prepared(last_committed+1);
#else //} }
void Replica::execute_prepared(bool committed) {
  //  if(!committed) return;
  if (last_tentative_execute < last_executed+1&& last_executed < last_stable
      +max_out&& !state.in_fetch_state() 
      && has_new_view()) {

    Pre_prepare *pp = prepared(last_executed+1);
#endif

    if (pp && pp->view() == view()) {
      // Can execute the requests in the message with sequence number
      // last_executed+1.
#ifdef SPECULATIVE
      last_tentative_execute = last_committed +1;
#else
      last_tentative_execute = last_executed+1;
#endif
      th_assert(pp->seqno() == last_tentative_execute, "Invalid execution");
			
      // Iterate over the requests in the message, calling execute for
      // each of them.
      Pre_prepare::Requests_iter iter(pp);
      Request req;
			
      while (iter.get(req)) {
	int cid = req.client_id();
	if (replies.req_id(cid) >= req.request_id()) {
	  // Request has already been executed and we have the reply to
	  // the request. Resend reply and don't execute request
	  // to ensure idempotence.
	  replies.send_reply(cid, view(), id());
	  continue;
	}
#ifdef THROUGHPUT_ADAPTIVE
        // increment request count only for "new" requests
        //printf("ADAPTIVE incrementing req_count\n");
        req_count++;
#endif
	// Obtain "in" and "out" buffers to call exec_command
	Byz_req inb;
	Byz_rep outb;
	Byz_buffer non_det;
	inb.contents = req.command(inb.size);
#ifdef SPECULATIVE
	outb.contents = replies.new_reply(cid, pp->history_digest(), 
					  pp->seqno(), outb.size);
#else
	outb.contents = replies.new_reply(cid, outb.size);
#endif
	non_det.contents = pp->choices(non_det.size);
				
	if (is_replica(cid)) {
	  // Handle recovery requests, i.e., requests from replicas,
	  // differently.  TODO: make more general to allow other types
	  // of requests from replicas.
	  //	  printf("\n\n\nExecuting recovery request seqno=%qd rep id=%d\n", last_tentative_execute, cid);

	  if (inb.size != sizeof(Seqno)) {
	    // Invalid recovery request.
	    continue;
	  }

	  // Change keys. TODO: could change key only for recovering replica.
	  if (cid != node_id){
	    printf("sending new key from execute spec/prep\n");
	    send_new_key();
	  }

	  // Store seqno of execution.	
	  max_rec_n = last_tentative_execute;

	  // Reply includes sequence number where request was executed.
	  outb.size = sizeof(last_tentative_execute);
	  memcpy(outb.contents, &last_tentative_execute, outb.size);
	} else {
	  // Execute command in a regular request.
	  //exec_command(&inb, &outb, &non_det, cid, false, exec_command_delay);
	  exec_command(&inb, &outb, &non_det, cid, false, exec_command_delay);
	  if (outb.size % ALIGNMENT_BYTES)
	    for (int i=0; i < ALIGNMENT_BYTES - (outb.size % ALIGNMENT_BYTES); i++)
	      outb.contents[outb.size+i] = 0;
	  //if (last_tentative_execute%100 == 0) 
	  //  printf("%s - %qd\n",((node_id == primary()) ? "P" : "B"), last_tentative_execute);
	}

	// Finish constructing the reply. 
	replies.end_reply(cid, req.request_id(), outb.size);

	// Send the reply. Replies to recovery requests are only sent
	// when the request is committed.
	if (outb.size != 0 && !is_replica(cid)) {
	  if (outb.size < 50 || req.replier() == node_id || req.replier() < 0) {
	    // Send full reply.
	    replies.send_reply(cid, view(), id(), !committed);
	  } else {
	    // Send empty reply.
#ifdef SPECULATIVE
	    Reply empty(view(), req.request_id(), node_id,
			replies.digest(cid), replies.history_digest(cid),
			last_committed, i_to_p(cid), !committed);
	    th_assert(0, "not sure about this last_executed for seqno");
#else
	    Reply empty(view(), req.request_id(), node_id,
			replies.digest(cid), i_to_p(cid), !committed);
	    send(&empty, cid);
#endif
	  }
	}
      }

#ifdef SPECULATIVE
      if ((last_committed+1) % checkpoint_interval == 0) {
        state.checkpoint(last_committed+1);
	if (status_messages_badly_needed && 
	    (last_committed > not_deprioritize_status_before_this +  max_out )) {
          printf("****** Deprioritizing status messages\n");
          status_messages_badly_needed = false;
	}

#else //}
      if ((last_executed+1) % checkpoint_interval == 0) {
	state.checkpoint(last_executed+1);

        if (status_messages_badly_needed && (last_executed > not_deprioritize_status_before_this + max_out )) {
	  printf("****** Deprioritizing status messages\n");
	  status_messages_badly_needed = false;
        }
#endif
#ifdef THROUGHPUT_ADAPTIVE
        Time current = rdtsc();
        float obs_throughput=0;
        checkpoints_in_new_view++;
        //printf("first_checkpoint_aftwe_view_change: %d, vc_already_triggered: %d, has_new_view(): %d\n", first_checkpoint_after_view_change, vc_already_triggered, has_new_view());
        if((checkpoints_in_new_view > 1) && has_new_view() && (id() != primary())) {
          //printf("inside\n");
          // when we perform the checkpoint, take a picture of our
          // throughput
          long long cp_duration = diffTime(current, last_cp_time);
          obs_throughput = ((float)req_count) / ((float)cp_duration);
          //printf("ADAPTIVE:  obs_throughput: %f, req_throughput: %f\n", obs_throughput, req_throughput);
          if (obs_throughput < req_throughput && !vc_already_triggered) {
            printf("ADAPTIVE view change! --- obs_throughput: %f, req_throughput: %f, increment: %f\n", obs_throughput, req_throughput, req_throughput_increment);
            vc_already_triggered = true;

            last_throughput_of_replica[primary()] = req_throughput-req_throughput_increment;
            float max_throughput = last_throughput_of_replica[0];
            for(int j = 1; j < num_replicas ; j++)
              max_throughput = MAX(max_throughput, last_throughput_of_replica[j]);
            req_throughput = max_throughput * new_req_throughput_scaling_factor; 
            req_throughput_increment = MAX(req_throughput * req_throughput_increment_scaling_factor, 0.000001);
            throughput_timer->restart();
          } 
          if ( obs_throughput > req_throughput && time_to_increment) {
            // increment the required throughput
            if ( (obs_throughput * new_req_throughput_scaling_factor) > (req_throughput + req_throughput_increment) ) {
              req_throughput = obs_throughput * new_req_throughput_scaling_factor;
              req_throughput_increment = MAX(req_throughput * req_throughput_increment_scaling_factor, 0.000001);
            } else {
              req_throughput += req_throughput_increment;
            }
            //last_throughput_of_replica[primary()] = obs_throughput;
          }
        }

        last_throughput_of_replica[primary()] = obs_throughput;

        if (first_checkpoint_after_view_change && has_new_view()){
          first_checkpoint_after_view_change = false;
          vc_already_triggered = false;
        } 
        req_count = 0;
        last_cp_time = current;
#endif
			}
		}
	}
}


  void Replica::execute_committed() {
    if (!state.in_fetch_state() && has_new_view()) {
      while (1) {
#ifdef SPECULATIVE
	if (last_committed >= last_stable + max_out ||
	    last_committed < last_stable)
	  return;
	Pre_prepare *pp = committed(last_committed + 1);

	if (pp && pp->view() == view()) {
          // Tentatively execute last_executed + 1 if needed.                                       
          execute_speculative(true);

          // Can execute the requests in the message with sequence number                           
          // last_executed+1.                                                                       
          last_committed = last_committed + 1;
          th_assert(pp->seqno() == last_committed, "Invalid execution");

#else
	if (last_executed >= last_stable+max_out || last_executed < last_stable)
	  return;
	Pre_prepare *pp = committed(last_executed+1);
	if (pp && pp->view() == view()) {
	  // Tentatively execute last_executed + 1 if needed.
	  execute_prepared(true);

	  // Can execute the requests in the message with sequence number
	  // last_executed+1.
	  last_executed = last_executed + 1;
	  th_assert(pp->seqno() == last_executed, "Invalid execution");

	  // Execute any buffered read-only requests
	  for (Request *m = ro_rqueue.remove(); m != 0; m
		 = ro_rqueue.remove()) {
	    execute_read_only(m);
	    delete m;
	  }
#endif
	  // Iterate over the requests in the message, marking the saved replies
	  // as committed (i.e., non-tentative for each of them).
	  Pre_prepare::Requests_iter iter(pp);
	  Request req;
	  while (iter.get(req)) {
	    int cid = req.client_id();
	    replies.commit_reply(cid);
	    if (is_replica(cid)) {
	      // Send committed reply to recovery request.
	      if (cid != node_id)
		replies.send_reply(cid, view(), id(), false);
	      else
		handle(replies.reply(cid)->copy(cid), true);
	    }
#ifdef FAIRNESS_ADAPTIVE
#ifdef SPECULATIVE
	    if ( (executed_snapshot[cid] != -1) &&
                 (executed_snapshot[cid] + fairness_bound < last_comitted) && has_new_view()) {
#else //}
	    if ( (executed_snapshot[cid] != -1) && 
		 (executed_snapshot[cid] + fairness_bound < last_executed) && has_new_view()) {
#endif
	      printf("ADAPTIVE: ****** Fairness violation!****** \n");
	      if(!vc_already_triggered){
		vc_already_triggered = true;
		throughput_timer->restart(); //starting throughput timer should work...
	      }
	    }
	    
	    executed_snapshot[cid] = -1;
#endif
	    // Remove the request from rqueue if present.
	    if (rqueue.remove(cid, req.request_id()))
	      vtimer->stop();
	  }
	  
#ifdef SPECULATIVE
	    Seqno execute_tmp = last_committed;
#else
	    Seqno execute_tmp = last_executed;
#endif	    

	  // Send and log Checkpoint message for the new state if needed.
	    if (execute_tmp % checkpoint_interval == 0) {
	    Digest d_state;
	    state.digest(execute_tmp, d_state);
	    Checkpoint *e = new Checkpoint(execute_tmp, d_state);
	    Certificate<Checkpoint> &cc = elog.fetch(execute_tmp);
	    cc.add_mine(e);
	    if (cc.is_complete()) {
	      mark_stable(execute_tmp, true);
	    }
	    send(e, All_replicas);
	    //	  printf(">>>>Checkpointing "); d_state.print(); printf(" <<<<\n"); fflush(stdout);
	  }
	  } else {
	  // No more requests to execute at this point.
	  break;
	}
      }

      if (rqueue.size() > 0) {
	if (primary() == node_id) {
	  // Send a pre-prepare with any buffered requests
	  send_pre_prepare();
	} else {
	  // If I am not the primary and have pending requests restart the
	  // timer.
	  //vtimer->start(); //necessary? useful?
	}
      }
    }
  }


void Replica::update_max_rec() {
  // Update max_rec_n to reflect new state.
  bool change_keys = false;
  for (int i=0; i < num_replicas; i++) {
    if (replies.reply(i)) {
      int len;
      char *buf = replies.reply(i)->reply(len);
      if (len == sizeof(Seqno)) {
	Seqno nr;
	memcpy(&nr, buf, sizeof(Seqno));
	
	if (nr > max_rec_n) {
	  max_rec_n = nr;
	  change_keys = true;
	}
	}
    }
  }
    
  // Change keys if state fetched reflects the execution of a new
  // recovery request.
  if (change_keys){
    printf("send new key from update_max_rec\n");
    send_new_key();
  }
}


void Replica::new_state(Seqno c) {   
  //  fprintf(stderr, ":n)e:w):s)t:a)t:e) ");
  printf("----- starting new state at %qd c with seqno %qd\n", c, seqno);
  if (vi.has_new_view(v) && c >= low_bound)
      has_nv_state = true;

#ifdef SPECULATIVE
  if (c > last_committed ){
    last_committed = last_tentative_execute = c;
#else // }
  if (c > last_executed) {
    last_executed = last_tentative_execute = c;
    //    fprintf(stderr, ":):):):):):):):) (new_state) Set le = %d\n", last_executed);
#endif
    if (replies.new_state(&rqueue))
      vtimer->stop();

    update_max_rec();

    if (c > last_prepared)
      last_prepared = c;

    if (c > last_stable+max_out) {
      mark_stable(c-max_out, elog.within_range(c-max_out) && elog.fetch(c-max_out).mine());
    }

    // Send checkpoint message for checkpoint "c"
    Digest d;
    state.digest(c, d);
    Checkpoint* ck = new Checkpoint(c, d);
    elog.fetch(c).add_mine(ck);
    //ck = elog.fetch(c).mine(); //not in PBFT
    //th_assert(ck, " I just added my checkpoint"); //not in PBFT
    send(ck, All_replicas);
  }

  // Check if c is known to be stable.
  int scount = 0;
  for (int i=0; i < num_replicas; i++) {
    Checkpoint* ck = sset.fetch(i);
    if (ck != 0 && ck->seqno() >= c) {
      th_assert(ck->stable(), "Invalid state");
      scount++;
    }
  }
  if (scount > f())
    mark_stable(c,true);

  if (c > seqno) {
    seqno = c;
  }

  // Execute any committed requests
  execute_committed();

#ifdef SPECULATIVE
  printf("new state has_nv_state %d has_new_view %d seqno %qd c %qd\n",
	 has_nv_state, has_new_view(), seqno, c);
  while (has_nv_state && has_new_view() && seqno > c){
    c++;
    printf("processing request c %qd and bumping it up to %qd \n", c, seqno);
    Prepared_cert &pc = plog.fetch(c);
    if (id() == primary()){
      Pre_prepare *pp = pc.my_pre_prepare();
      replies.append_digest(replies.history_digest(), pp->digest());
    }else{
      Prepare *p = pc.my_prepare();
      replies.append_digest(replies.history_digest(), p->digest());
    }
  }
#else
  // Execute any buffered read-only requests
  for (Request *m = ro_rqueue.remove(); m != 0; m = ro_rqueue.remove()) {
    execute_read_only(m);
    delete m;
  }
#endif

  if (rqueue.size() > 0) {
    if (primary() == id()){
      // Send pre-prepares for any buffered requests
      send_pre_prepare();
    } else 
      //printf("vtimer restart (2)\n");
      vtimer->restart();
  }  

#ifdef SPECULATIVE
  printf("------New state completed @ seqno %qd @ c %qd last_exec %qd and history ", 
	 seqno,c, last_committed);
  replies.history_digest().print();printf("\n");
#endif
}



void Replica::mark_stable(Seqno n, bool have_state) {
  //XXXXXcheck if this should be < or <=

  //  fprintf(stderr, "mark stable n %qd laststable %qd\n", n, last_stable);
  if (n <= last_stable)
    return;

  last_stable = n;
  if (last_stable > low_bound) {
    low_bound = last_stable;
  }

#ifdef SPECULATIVE
  if (have_state && last_stable > last_committed) {
    last_committed = last_tentative_execute = last_stable;
#else  // }
  if (have_state && last_stable > last_executed) {
    last_executed = last_tentative_execute = last_stable;
#endif
    //    fprintf(stderr, ":):):):):):):):) (mark_stable) Set le = %d\n", last_executed);
    replies.new_state(&rqueue);
    update_max_rec();
   
    if (last_stable > last_prepared)
      last_prepared = last_stable;
  }
  //  else
  //    fprintf(stderr, "OH BASE! OH CLU!\n");

  if (last_stable > seqno)
    seqno = last_stable;

  //  fprintf(stderr, "mark_stable: Truncating plog to %ld have_state=%d\n", last_stable+1, have_state);
  plog.truncate(last_stable+1);
  clog.truncate(last_stable+1);
  vi.mark_stable(last_stable);
  elog.truncate(last_stable);
#ifdef SPECULATIVE
  state.discard_checkpoint(last_stable, last_committed);
#else
  state.discard_checkpoint(last_stable, last_executed);
#endif
  brt.mark_stable(last_stable);

  if (have_state) {
    // Re-authenticate my checkpoint message to mark it as stable or
    // if I do not have one put one in and make the corresponding
    // certificate complete.
    Checkpoint *c = elog.fetch(last_stable).mine();
    if (c == 0) {
      Digest d_state;
      state.digest(last_stable, d_state);
      c = new Checkpoint(last_stable, d_state, true);
      elog.fetch(last_stable).add_mine(c);
      elog.fetch(last_stable).make_complete();
    } else {
      c->re_authenticate(0,true);
    }
  
  }

  // Go over sset transfering any checkpoints that are now within
  // my window to elog.
  Seqno new_ls = last_stable;
  for (int i=0; i < num_replicas; i++) {
    Checkpoint* c = sset.fetch(i);
    if (c != 0) {
      Seqno cn = c->seqno();
      if (cn < last_stable) {
	c = sset.remove(i);
	delete c;
	continue;
      }

      if (cn <= last_stable + max_out) {
	Certificate<Checkpoint>& cs = elog.fetch(cn);
	cs.add(sset.remove(i));
	if (cs.is_complete() && cn > new_ls) 
	  new_ls = cn;
      }
    }
  }
  
  //XXXXXXcheck if this is safe.
  if (new_ls > last_stable) {
    //    fprintf(stderr, "@@@@@@@@@@@@@@@               @@@@@@@@@@@@@@@               @@@@@@@@@@@@@@@\n");
    mark_stable(new_ls, elog.within_range(new_ls) && elog.fetch(new_ls).mine());
  }
  
  // Try to send any Pre_prepares for any buffered requests.
  if (primary() == id()) 
    send_pre_prepare();
}


void Replica::handle(Data *m) {
  //printf("\t\t\t\t\t\tdata\n");
  //printf("received data\n");
  state.handle(m);
}


void Replica::handle(Meta_data *m) {
  //printf("\t\t\t\t\t\tmetadata\n");

  state.handle(m);
}


void Replica::handle(Meta_data_d *m) {
  //printf("\t\t\t\t\t\tmetadatad\n");

  state.handle(m);
}


void Replica::handle(Fetch *m) {
  //printf("\t\t\t\t\t\tfetch\n");

  int mid = m->id();
  if (!state.handle(m, last_stable) && last_new_key != 0) {
    send(last_new_key, mid);
  }
}

void Replica::send_new_key() {
  Node::send_new_key();
  printf(" call to send new key\n");
  // Cleanup messages in incomplete certificates that are
  // authenticated with the old keys.
  int max = last_stable+max_out;
  int min = last_stable+1;
  for (Seqno n=min; n <= max; n++) {
    if (n%checkpoint_interval == 0) 
      elog.fetch(n).mark_stale();
  }

#ifdef SPECULATIVE
  if (last_committed > last_stable)
    min = last_stable+1;
#else
  if (last_executed > last_stable)
    min = last_executed+1;
#endif
  for (Seqno n=min; n <= max; n++) {
    plog.fetch(n).mark_stale();
    clog.fetch(n).mark_stale();
  }

  vi.mark_stale();
  state.mark_stale();
}


void Replica::send_status(bool force) {
  // Check how long ago we sent the last status message.
  Time cur = currentTime();
  if (force || diffTime(cur, last_status) > 100000) {
  //if(1) {
    // Only send new status message if last one was sent more
    // than 100 milliseconds ago.
    // [elwong] ... or if we are forcing it.
    last_status = cur;

    if (rr) {
      // Retransmit recovery request if I am waiting for one.
      send(rr, All_replicas);
    }
    
    // If fetching state, resend last fetch message instead of status.
    if (state.retrans_fetch(cur)) {
      state.send_fetch(true);
      return;
    }

#ifdef SPECULATIVE
    Status s(v, last_stable, last_committed, has_new_view(), vi.has_nv_message(v));
#else
    Status s(v, last_stable, last_executed, has_new_view(), vi.has_nv_message(v));
#endif

    if (has_new_view()) {
      // Set prepared and committed bitmaps correctly
      int max = last_stable+max_out;
#ifdef SPECULATIVE
      for (Seqno n = last_committed+1; n <= max; n++) {
#else//}
      for (Seqno n=last_executed+1; n <= max; n++) {
#endif
	Prepared_cert& pc = plog.fetch(n);
	if (pc.is_complete() ) { 
	  s.mark_prepared(n);
	  if (clog.fetch(n).is_complete() ) {
	    s.mark_committed(n);
	  }
	} else {
	  // Ask for missing big requests
	  if (!pc.is_pp_complete() && pc.pre_prepare() && pc.num_correct() >= f())
	    s.add_breqs(n, pc.missing_reqs());
	}
      }
    } else {
      vi.set_received_vcs(&s);
      vi.set_missing_pps(&s);
    }
  
    // Multicast status to all replicas.
    s.authenticate();
    send(&s, All_replicas);
  }
}

void Replica::handle(Reply *m, bool mine) {
//  th_assert(false,"should not be there...\n");
  int mid = m->id();
  int mv = m->view();

#ifdef SPECULATIVE
  if (rr && rr->request_id() == m->request_id() && (mine || !m->is_speculative())){
#else//}
  if (rr && rr->request_id() == m->request_id() && (mine || !m->is_tentative())) {
#endif
    // Only accept recovery request replies that are not tentative.
    bool added = (mine) ? rr_reps.add_mine(m) : rr_reps.add(m);
    if (added) {
      if (rr_views[mid] < mv)
	rr_views[mid] = mv;

      if (rr_reps.is_complete()) {
	// I have a valid reply to my outstanding recovery request.
	// Update recovery point
	int len;
	const char *rep = rr_reps.cvalue()->reply(len);
	th_assert(len == sizeof(Seqno), "Invalid message");

	Seqno rec_seqno;
	memcpy(&rec_seqno, rep, len);
	Seqno new_rp = rec_seqno/checkpoint_interval * checkpoint_interval + max_out;
	if (new_rp > recovery_point)
	  recovery_point = new_rp;

	//	printf("XXX Complete rec reply with seqno %qd rec_point=%qd\n",rec_seqno,  recovery_point);	
	
	// Update view number
	//View rec_view = K_max<View>(f()+1, rr_views, n(), View_max);
	delete rr;
	rr = 0;
      }
    }
    return;
  }
  delete m;
}


void Replica::send_null() {
  //printf("@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@\n");
  th_assert(id() == primary(), "Invalid state");

  Seqno max_rec_point = max_out
    +(max_rec_n+checkpoint_interval-1)/checkpoint_interval*checkpoint_interval;

  if (max_rec_n && max_rec_point > last_stable && has_new_view()) {
#ifdef SPECULATIVE
    if (rqueue.size() == 0 && seqno <= last_committed && seqno+1 <= max_out+last_stable) {
#else //}
    if (rqueue.size() == 0 && seqno <= last_executed && seqno+1 <= max_out+last_stable) {
#endif
      // Send null request if there is a recovery in progress and there
      // are no outstanding requests.
      seqno++;
      Req_queue empty;
#ifdef SPECULATIVE
      Pre_prepare *pp = new Pre_prepare(view(), seqno, empty, replies.history_digest());
#else
      Pre_prepare* pp = new Pre_prepare(view(), seqno, empty);
#endif
      send(pp, All_replicas);
      plog.fetch(seqno).add_mine(pp);
    }
  }
  ntimer->restart();
  // TODO: backups should force view change if primary does not send null requests
  // to allow recoveries to complete.
}


//
// Timeout handlers:
//

void vtimer_handler() {
  //printf(" ###### @@ vtimer expired! @@ ######\n");

  th_assert(replica, "replica is not initialized\n");
  if (!replica->delay_vc())
  {
//    printf("sending view change (vtimer expired)\n");
    replica->send_view_change();
  } 
  else 
  {
    //printf("restarting vtimer 4\n");
    //printf("vtimer restart (3)\n");
    replica->vtimer->restart();
  }
}

#ifdef THROUGHPUT_ADAPTIVE
//timer started after having observed a throughput lower than expected
void throughput_timer_handler() {
  th_assert(replica, "replica is not initialized\n");
  if (!replica->delay_vc() ) {
//    printf("ADAPTIVE: sending view change for view %qd (throughput_timer expired)\n", replica->view()+1);
#ifdef DELAY_ADAPTIVE
    replica->pre_prepare_timer->stop();
#endif
    replica->send_view_change();
  } else {
    if(replica->delay_vc()) {
//      printf("ADAPTIVE: restarting throughput_timer \n");
      replica->throughput_timer->restart();
    }
  }
}

void increment_timer_handler() {
  th_assert(replica,"replica is not initialized\n");
  replica->time_to_increment = true;
//  printf("--- time to increment the required throughput! ---\n");
}
#endif


#ifdef DELAY_ADAPTIVE
void pre_prepare_timer_handler() {
  th_assert(replica, "replica is not initialized\n");

  if( replica->id() != replica->primary() && replica->received_pre_prepares < expected_pre_prepares) {
      if (!replica->delay_vc()){
#ifdef THROUGHPUT_ADAPTIVE
        replica->throughput_timer->restop();
#endif
        printf("ADAPTIVE: sending view change for view %qd (pre_prepare_timer expired)\n", replica->view()+1);
        replica->send_view_change();
	cur_pre_prepare_timer_duration *= 2;
	replica->pre_prepare_timer->adjust(cur_pre_prepare_timer_duration);
      } else {
        replica->pre_prepare_timer->restart();
      }
  } else {
    replica->pre_prepare_timer->restop();
//    printf("we passed the pre-prepare timer check, now we're restarting it\n");
    cur_pre_prepare_timer_duration = pre_prepare_timer_duration ;
    replica->pre_prepare_timer->adjust(cur_pre_prepare_timer_duration);
    replica->pre_prepare_timer->restart();
    replica->received_pre_prepares = 0;
  }
}

#endif


 void delay_pre_prepare_timer_handler(){
   if (replica->has_new_view() && replica->primary() == replica->id())
     replica->send_pre_prepare(true);
 }

void stimer_handler() {
  //printf("--- stimer expired --- \n");
  th_assert(replica, "replica is not initialized\n");
  replica->send_status();
  replica->stimer->restart();
  //printf("--- stimer restarted --- \n");
}


void ntimer_handler() {
  //printf(" --- ntimer expired --- \n");
  th_assert(replica, "replica is not initialized\n");
  replica->send_null();
}


bool Replica::has_req(int cid, const Digest &d) {
  Request* req = rqueue.first_client(cid);

  if (req && req->digest() == d)
    return true;

  return false;
}

Message* Replica::pick_next_status() {
  for(int identities_checked=0; identities_checked < num_replicas; identities_checked++) {
    status_to_process++;
    status_to_process %= num_replicas;
    if (status_pending[status_to_process] != NULL){
      //printf("@@@@@@ handling stored status from %d @@@@@@ \n",status_to_process);
      Message* m = status_pending[status_to_process];
      status_pending[status_to_process] = NULL;
      return m;
    } 
  }
  // status_pending is empty...
  return NULL; 
}

bool Replica::cb_write_message(Message* message){

  //printf("Writing --- ri: %d, wi: %d \n",cb_read_index, cb_write_index);
  if ( ((cb_write_index + 1) % circular_buffer_size) == cb_read_index)
    return false; //the buffer is full

  //the buffer is not full
  circular_buffer[cb_write_index] = message;
  cb_write_index = (cb_write_index+1) % circular_buffer_size;
  return true;
}

Message* Replica::cb_read_message(){

  //printf("Reading --- ri: %d, wi: %d \n",cb_read_index, cb_write_index);
  if (cb_write_index == cb_read_index)
    return NULL; //the buffer is empty

  //the buffer is not empty
  Message* temp = circular_buffer[cb_read_index];
  circular_buffer[cb_read_index] = circular_buffer_magic;
  cb_read_index = (cb_read_index + 1) % circular_buffer_size;
  return temp;
}
