#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <pthread.h>
#include "Wrapped_request.h"
#include "Verifier_thread.h"

void *Verifier_thread_startup(void *_tgtObject) {
  Verifier_thread *tgtObject = (Verifier_thread *)_tgtObject;
//  printf("Running a verifier thread object in a new thread\n");
  void *threadResult = tgtObject->run();
//  printf("Deleting object\n");
  delete tgtObject;
  return threadResult;
}


void *Verifier_thread::run(void)
{
  while(!replica){
    printf("replica not initialized yet...\n");
    sleep(1);
  }
  printf("%s\n",message);

  //blacklist malloc and initialization
  blacklisted = (bool*)malloc(sizeof(bool) * node->num_principals);
  for(int i=0; i< node->num_principals; i++)
    blacklisted[i]=false;

  //DEBUG: set blacklisted 10 = true, just to see if it works...
  //blacklisted[10]=true;	

  // initialize, bind and set the client socket
  // (used to receive request from clients)

  client_socket = socket(AF_INET, SOCK_DGRAM, 0);

  client_address.sin_family = AF_INET;
  //client_address.sin_addr.s_addr = node->principals[replica->id()]->address()->sin_addr.s_addr;
  client_address.sin_addr.s_addr = htonl(INADDR_ANY);
  client_address.sin_port = client_port;

//  printf("replica %d, binding client socket on port %d\n",replica->id(), client_address.sin_port);

  error = bind(client_socket, (struct sockaddr*)&client_address, sizeof(Addr));
  if (error < 0) {
    perror("Unable to name client socket");
    exit(1);
  }

  printf("WARNING: async sockets\n");
  error = fcntl(client_socket, F_SETFL, O_NONBLOCK);
  if (error < 0) {
    perror("unable to set socket to asynchronous mode");
    exit(1);
  }

#define LARGE_SND_BUFF
#ifdef LARGE_SND_BUFF
    error = setsockopt(client_socket, SOL_SOCKET, SO_SNDBUF,
                      (char*)&snd_buf_size, sizeof(snd_buf_size));
    if (error < 0) {
      perror("unable to increase send buffer size");
      exit(1);
    }
#endif


#define LARGE_RCV_BUFF
#ifdef LARGE_RCV_BUFF
    error = setsockopt(client_socket, SOL_SOCKET, SO_RCVBUF,
                        (char*)&rcv_buf_size, sizeof(rcv_buf_size));
    if (error < 0) {
      perror("unable to increase send buffer size");
      exit(1);
    }
#endif

#ifdef NO_UDP_CHECKSUM
    error = setsockopt(client_socket, SOL_SOCKET, SO_NO_CHECK,
                                       (char*)&no_check, sizeof(no_check));
    if (error < 0) {
      perror("unable to turn of UDP checksumming");
      exit(1);
    }
#endif //NO_UDP_CHECKSUM

  // initialize, bind and set the server socket
  // (used to send verified requests to replica's main thread)

  server_socket = socket(AF_INET, SOCK_DGRAM, 0);

  server_address.sin_family = AF_INET;
  server_address.sin_addr.s_addr = inet_addr("127.0.0.1"); //only for communications on the same physical host!
  server_address.sin_port = htons(node->principals[node->node_id]->get_verifier_thread_bind_port()); //main thread is listening on port 9876

  error = bind(server_socket, (struct sockaddr*)&server_address, sizeof(Addr));
  if (error < 0) {
    perror("Unable to name server socket");
    exit(1);
  }

  printf("WARNING: async sockets\n");
  error = fcntl(server_socket, F_SETFL, O_NONBLOCK);
  if (error < 0) {
    perror("unable to set socket to asynchronous mode");
    exit(1);
  }

#define LARGE_SND_BUFF
#ifdef LARGE_SND_BUFF
  error = setsockopt(server_socket, SOL_SOCKET, SO_SNDBUF,
                    (char*)&snd_buf_size, sizeof(snd_buf_size));
  if (error < 0) {
    perror("unable to increase send buffer size");
    exit(1);
  }
#endif


#define LARGE_RCV_BUFF
#ifdef LARGE_RCV_BUFF
  error = setsockopt(server_socket, SOL_SOCKET, SO_RCVBUF,
                        (char*)&rcv_buf_size, sizeof(rcv_buf_size));
  if (error < 0) {
    perror("unable to increase send buffer size");
    exit(1);
  }
#endif

#ifdef NO_UDP_CHECKSUM
  error = setsockopt(server_socket, SOL_SOCKET, SO_NO_CHECK,
                                      (char*)&no_check, sizeof(no_check));
  if (error < 0) {
    perror("unable to turn of UDP checksumming");
    exit(1);
  }
#endif //NO_UDP_CHECKSUM

  //sockets ready...
  //initialize the destination address

  dest_address.sin_family = AF_INET;
  dest_address.sin_addr.s_addr = inet_addr("127.0.0.1"); //only for communications on the same physical host!
  dest_address.sin_port = htons(node->principals[node->node_id]->get_main_thread_bind_port()); //main thread is listening on port 9876
  
  //start the main processing loop

 

  while(1) {
    
    select_timeout.tv_sec = 0;
    select_timeout.tv_usec = 10000000; //10ms, probably it could even be infinite...  
   
    FD_ZERO(&fdset);
    FD_SET(client_socket, &fdset);

    ret = select(client_socket+1, &fdset, 0, 0, &select_timeout);

    if (ret > 0 && FD_ISSET(client_socket, &fdset)){
      //something to read from clients!!
//      printf("------------------- Verifier thread: something ready to read...\n");
      //initializing m
      m = new Message(Max_message_size);

      //reading from socket and filling m
      ret = recvfrom(client_socket, m->contents(), m->msize(), 0, 0, 0);

        
      if (ret >= (int)sizeof(Message_rep) && ret >= m->size()) {


        if( (m->tag() < 1) || (m->tag() > 15) && (m->tag() != 18) ){
          delete m;
          continue;
        }



        if(m->tag() == Wrapped_request_tag){
          Wrapped_request* wrapped= (Wrapped_request*)m;
//          printf("Wrap size: %d\n",wrapped->size()); 

	  //check if m has been sent by a blacklisted client.
	  //if this is the case, just drop the request and continue.
	  if (blacklisted[wrapped->client_id()]) {
            delete m;
            continue;
	  }	

          bool verified_wrap = wrapped->verify_MAC();
          if (!verified_wrap){
            delete m;
            continue;
          }
	
	  // verify_MAC returned true, so digest and MAC are ok. 
	  // now we call verify_request to verify the signature.

	  verified_wrap = wrapped->verify_request();
	  if (!verified_wrap){
            // adding the sender of this wrapped request to the blacklist
	    //blacklisted[wrapped->client_id()] = true;
	    printf("*** +++ --- \\\\\\ Blacklisted client %d /// --- +++ ***\n",wrapped->client_id());
	    delete m;
  	    continue;
	  }

	  // the request also verified. we can add the request.

          Wrapped_request::Requests_iter iter(wrapped);
          Request req;
	  iter.get(req);

 /* 
	  // potentially unsafe multi threaded access to memory :(
	  // if the message is in the past then dont verify
	  Request_id last_rid = replica->replies.req_id(req.client_id());
	  if (last_rid == req.request_id() && req.client_id() == wrapped->client_id()){
	    ; // its an old request, we just need to process it for retransmission purposes
	    ; // we dont need to verify the signature
	  } else if (last_rid < req.request_id()){
	    // its the next request, we need to verify it
	    if (!req.verify()){
	      ; // need to blacklist the sender of the wrapped message
	      delete m;
	      continue;
	    }
	  } else{
	    // throw it out now.  its not for retransmitting and its
	    // nto a new request to process, so just throw it out and dont bother verifying it
	    delete m;
	    continue;
	  }
	    
*/

          Request* to_write;
          to_write = req.clone();
          //sendto(server_socket, req.contents(),req.size(),0,(struct sockaddr*)&dest_address, sizeof(Addr));
          if (!replica->cb_write_message((Message*)to_write)){
            //printf("Unable to write request in the buffer \n");
            delete to_write;
          }
          delete m;
          continue;
        }

        if(m->tag() == New_key_tag){
          ((New_key*)m)->verify(); 
        }

        //sending the message to the main thread
        if (!replica->cb_write_message(m)){
          //printf("Unable to write request in the buffer \n");
          delete m;
        }

      } else {
        delete m;
      } 
    }
  }
  return NULL;
}

